"""Misty2py is a Python package for Misty II development using [Misty's REST API](https://docs.mistyrobotics.com/misty-ii/rest-api/api-reference/).

## Features
Misty2py can be used to:

- **perform actions** via sending a `POST` or `DELETE` requests to Misty's API;
- **obtain information** via sending a `GET` request to Misty's API;
- **receive streams of data** via subscribing to event types on Misty's Websockets API.

Misty2py uses following concepts:

- **action keywords** - keywords for endpoints of Misty's API that correspond to performing actions;
- **information keywords** - keywords for endpoints of Misty's API that correspond to retrieving information;
- **data shortcuts** - keywords for commonly used data that is supplied to Misty's API as the body of a `POST` request.

## Installation

### Poetry

To install misty2py, run `pip install misty2py`.

### From source

- If this is your first time using `misty2py` from source, do following:

  - Get Poetry (`python -m pip install poetry`) if you do not have it yet.
  - Copy `.env.example` to `.env`.
  - Replace the placeholder values in the new `.env` file.
  - Run `poetry install` to obtain all dependencies.

- Run the desired script via `poetry run python -m [name]` where `[name]` is the placeholder for the module location (in Python notation).
- If the scripts run but your Misty does not seem to respond, you have most likely provided an incorrect IP address for `MISTY_IP_ADDRESS` in `.env`.
- Pytests can be run via `poetry run pytest .`.
- The coverage report can be obtained via `poetry run pytest --cov-report html --cov=misty2py tests` for HTML output or via `poetry run pytest --cov=misty2py tests` for terminal output.

## Contribute

This project is not currently open for contributions. However, you can report issues via the [Issues Tracker](https://github.com/ChrisScarred/misty2py/issues) and inspect the [Source Code](https://github.com/ChrisScarred/misty2py).

## License

This project is licensed under the MIT License.

## Usage

### Getting started

- Start by making **a new instance** of `misty2py.robot`'s `Misty` by `misty_robot = Misty("ip_address_here")`
  - Substitute `ip_address_here` with the IP address of your Misty.
  - `misty2py.utils.env_loader` module contains `EnvLoader` class that can be used to load Misty's IP from the .env file in your project's home directory:
    - Create the `.env` file and write `MISTY_IP_ADDRESS="[ip_address_here]"` in it.
    - Initialise an `EnvLoader` object via `env_loader = EnvLoader()`; this loads the environment variables.
    - Use `env_loader.get_ip()` to obtain the IP address.
- Use the method `misty_robot.perform_action()` to tell Misty to **perform an action**.
- Use the method `misty_robot.get_info()` to tell Misty to **return information**.
- Use the method `misty_robot.event()` to initialise, obtain and stop **continuous streams of data** from Misty's event types.

### Obtaining information

Obtaining digital information is handled by `misty2py.robot::get_info` method.

`misty2py.robot::get_info` has following arguments:

- `info_name` - *required;* the string information keyword corresponding to an endpoint in Misty's API;
- `params` - *optional;* a dictionary of parameter name and parameter value pairs, defaults to `{}`.

### Performing actions

Performing physical and digital actions including removal of non-system files is handled by `misty2py.robot::perform_action()` method.

`misty2py.robot::perform_action()` has following arguments:

- `action_name` - *required;* the string action keyword corresponding to an endpoint in Misty's API;
- `data` - *optional;* the data to pass to the request as a dictionary or a data shortcut (string), defaults to `{}`.

### Event types

To obtain event data in Misty's framework, it is required to **subscribe** to an event type on Misty's websocket server. Misty's websocket server then streams data to the websocket client, in this implementation via a separate thread. To **access this data,** `misty2py.robot::event` method must be called with `"get_data"` parameter from the main thread. When data are no longer required to be streamed to the client, an event type can be **unsubscribed** to kill the event thread.

#### Subscription

Subscribe to an event via `misty2py.robot::event` with the parameter `"subscribe"` and following keyword arguments:

- `type` - *required;* event type string as documented in [Event Types Docs](https://docs.mistyrobotics.com/misty-ii/robot/sensor-data/ "Misty Robotics Event Types").
- `name` - *optional;* a custom event name string; must be unique.
- `return_property` - *optional;* the property to return from Misty's websockets; all properties are returned if return_property is not supplied.
- `debounce` - *optional;* the interval in ms at which new information is sent; defaults to 250.
- `len_data_entries` - *optional;* the maximum number of data entries to keep (discards in fifo style); defaults to 10.
- `event_emitter` - *optional;* an event emitter function which emits an event upon message recieval. Supplies the message content as an argument.

#### Accessing the data and the log

Access the data of an event or its log via `misty2py.robot::event` with the parameter `"get_data"` or `"get_log"` and a keyword argument `name` (the name of the event).

#### Unsubscribing

Unsubscribe from an event via `misty2py.robot::event` with the parameter `"unsubscribe"` and a keyword argument `name` (the name of the event).

#### Basic events usage example

```python
import time

from misty2py.robot import Misty
from misty2py.utils.env_loader import EnvLoader

env_loader = EnvLoader

m = Misty(env_loader.get_ip())

d = m.event("subscribe", type = "BatteryCharge")
e_name = d.get("event_name")

time.sleep(1)

d = m.event("get_data", name = e_name)

d = m.event("unsubscribe", name = e_name)
```

#### Event emitter usage example

```python
import time
from pymitter import EventEmitter

from misty2py.robot import Misty
from misty2py.utils.env_loader import EnvLoader

env_loader = EnvLoader

m = Misty(env_loader.get_ip())
ee = EventEmitter()
event_name = "myevent_001"

@ee.on(event_name)
def listener(data):
    print(data)

d = m.event("subscribe", type = "BatteryCharge", 
            name = event_name, event_emitter = ee)

time.sleep(2)

d = m.event("unsubscribe", name = event_name)
```

### Adding custom keywords and shortcuts

Custom keywords and shortcuts can be passed to a Misty object while declaring a new instance by using the optional arguments:

- `custom_info` for custom information keywords (a dictionary with keys being the information keywords and values being the endpoints),
- `custom_actions` for custom action keywords (a dictionary with keys being the action keywords and values being a dictionary `{"endpoint" : "edpoint_value", "method" : "method_value"}` where `method_value` is either `post` or `delete`),
- `custom_data` for custom data shortcuts (a dictionary with keys being the data shortcuts and values being the dictionary of data values).

An example:

```python
custom_allowed_infos = {
    "hazards_settings": "api/hazards/settings"
}

custom_allowed_data = {
    "amazement": {
        "FileName": "s_Amazement.wav"
    },
    "red": {
        "red": "255",
        "green": "0",
        "blue": "0"
    }
}

custom_allowed_actions = {
    "audio_play" : {
        "endpoint" : "api/audio/play",
        "method" : "post"
    },
    "delete_audio" : {
        "endpoint" : "api/audio",
        "method" : "delete"
    }
}

misty_robot = Misty("0.0.0.0", 
    custom_info=custom_allowed_infos, 
    custom_actions=custom_allowed_actions, 
    custom_data=custom_allowed_data)
```


## Default Keywords and Shortcuts

<details> 
  <summary>List of supported action keywords</summary>

<ul>
<li><code>led</code> for <strong>post</strong> request to <code>api/led</code> endpoint</li>
<li><code>led_trans</code> for <strong>post</strong> request to <code>api/led/transition</code> endpoint</li>
<li><code>notification_settings</code> for <strong>post</strong> request to <code>api/notification/settings</code> endpoint</li>
<li><code>audio_upload</code> for <strong>post</strong> request <code>api/audio</code> to endpoint</li>
<li><code>audio_play</code> for <strong>post</strong> request to <code>api/audio/play</code> endpoint</li>
<li><code>audio_pause</code> for <strong>post</strong> request to <code>api/audio/pause</code> endpoint</li>
<li><code>audio_stop</code> for <strong>post</strong> request to <code>api/audio/stop</code> endpoint</li>
<li><code>audio_delete</code> for <strong>delete</strong> request to <code>api/audio</code> endpoint</li>
<li><code>audio_record_start</code> for <strong>post</strong> request to <code>api/audio/record/start</code> endpoint</li>
<li><code>audio_record_stop</code> for <strong>post</strong> request to <code>api/audio/record/stop</code> endpoint</li>
<li><code>audio_disable</code> for <strong>post</strong> request to <code>api/services/audio/disable</code> endpoint</li>
<li><code>audio_enable</code> for <strong>post</strong> request to <code>api/services/audio/enable</code> endpoint</li>
<li><code>image_upload</code> for <strong>post</strong> request to <code>api/images</code> endpoint</li>
<li><code>image_show</code> for <strong>post</strong> request to <code>api/images/display</code> endpoint</li>
<li><code>image_settings</code> for <strong>post</strong> request to <code>api/images/settings</code> endpoint</li>
<li><code>image_delete</code> for <strong>delete</strong> request to <code>api/images</code> endpoint</li>
<li><code>text_show</code> for <strong>post</strong> request to <code>api/text/display</code> endpoint</li>
<li><code>text_settings</code> for <strong>post</strong> request to <code>api/text/settings</code> endpoint</li>
<li><code>video_upload</code> for <strong>post</strong> request to <code>api/videos</code> endpoint</li>
<li><code>video_show</code> for <strong>post</strong> request to <code>api/videos/display</code> endpoint</li>
<li><code>video_settings</code> for <strong>post</strong> request to <code>api/videos/settings</code> endpoint</li>
<li><code>video_delete</code> for <strong>delete</strong> request to <code>api/videos</code> endpoint</li>
<li><code>blink_mapping_delete</code> for <strong>delete</strong> request to <code>api/blink/images</code> endpoint</li>
<li><code>slam_enable</code> for <strong>post</strong> request to <code>api/services/slam/enable</code> endpoint</li>
<li><code>slam_disable</code> for <strong>post</strong> request to <code>api/services/slam/disable</code> endpoint</li>
<li><code>slam_sensors_reset</code> for <strong>post</strong> request to <code>api/slam/reset</code> endpoint</li>
<li><code>slam_mapping_start</code> for <strong>post</strong> request to <code>api/slam/map/start</code> endpoint</li>
<li><code>slam_mapping_stop</code> for <strong>post</strong> request to <code>api/slam/map/stop</code> endpoint</li>
<li><code>slam_map_current</code> for <strong>post</strong> request to <code>api/slam/map/current</code> endpoint</li>
<li><code>slam_map_rename</code> for <strong>post</strong> request to <code>api/slam/map/rename</code> endpoint</li>
<li><code>slam_infrared_settings</code> for <strong>post</strong> request to <code>api/slam/settings/ir</code> endpoint</li>
<li><code>slam_visible_settings</code> for <strong>post</strong> request to <code>api/slam/settings/visible</code> endpoint</li>
<li><code>slam_map_delete</code> for <strong>delete</strong> request to <code>api/slam/map</code> endpoint</li>
<li><code>slam_docking_locate_start</code> for <strong>post</strong> request to <code>api/slam/docking/start</code> endpoint</li>
<li><code>slam_docking_locate_stop</code> for <strong>post</strong> request to <code>api/slam/docking/stop</code> endpoint</li>
<li><code>streaming_slam_start</code> for <strong>post</strong> request to <code>api/slam/streaming/start</code> endpoint</li>
<li><code>streaming_slam_stop</code> for <strong>post</strong> request to <code>api/slam/streaming/stop</code> endpoint</li>
<li><code>slam_track_start</code> for <strong>post</strong> request to <code>api/slam/track/start</code> endpoint</li>
<li><code>slam_track_stop</code> for <strong>post</strong> request to <code>api/slam/track/stop</code> endpoint</li>
<li><code>recording_start</code> for <strong>post</strong> request to <code>api/videos/recordings/start</code> endpoint</li>
<li><code>recording_stop</code> for <strong>post</strong> request to <code>api/videos/recordings/stop</code> endpoint</li>
<li><code>recording_rename</code> for <strong>post</strong> request to <code>api/videos/recordings/rename</code> endpoint</li>
<li><code>recording_delete</code> for <strong>delete</strong> request to <code>api/videos/recordings</code> endpoint</li>
<li><code>face_detection_start</code> for <strong>post</strong> request to <code>api/faces/detection/start</code> endpoint</li>
<li><code>face_detection_stop</code> for <strong>post</strong> request to <code>api/faces/detection/stop</code> endpoint</li>
<li><code>face_recognition_start</code> for <strong>post</strong> request to <code>api/faces/recognition/start</code> endpoint</li>
<li><code>face_recognition_stop</code> for <strong>post</strong> request to <code>api/faces/recognition/stop</code> endpoint</li>
<li><code>face_train_start</code> for <strong>post</strong> request to <code>api/faces/training/start</code> endpoint</li>
<li><code>face_train_cancel</code> for <strong>post</strong> request to <code>api/faces/training/cancel</code> endpoint</li>
<li><code>face_delete</code> for <strong>delete</strong> request to <code>api/faces</code> endpoint</li>
<li><code>skill_upload</code> for <strong>post</strong> request to <code>api/skills</code> endpoint</li>
<li><code>skill_start</code> for <strong>post</strong> request to <code>api/skills/start</code> endpoint</li>
<li><code>skills_reload</code> for <strong>post</strong> request to <code>api/skills/reload</code> endpoint</li>
<li><code>skill_load</code> for <strong>post</strong> request to <code>api/skills/load</code> endpoint</li>
<li><code>skill_cancel</code> for <strong>post</strong> request to <code>api/skills/cancel</code> endpoint</li>
<li><code>skill_delete</code> for <strong>delete</strong> request to <code>api/skills</code> endpoint</li>
<li><code>wifi_add</code> for <strong>post</strong> request to <code>api/networks/create</code> endpoint</li>
<li><code>wifi_connect</code> for <strong>post</strong> request to <code>api/networks</code> endpoint</li>
<li><code>wifi_delete</code> for <strong>delete</strong> request to <code>api/networks</code> endpoint</li>
<li><code>wifi_hotspot_start</code> for <strong>post</strong> request to <code>api/networks/hotspot/start</code> endpoint</li>
<li><code>wifi_hotspot_stop</code> for <strong>post</strong> request to <code>api/networks/hotspot/stop</code> endpoint</li>
<li><code>write_serial</code> for <strong>post</strong> request to <code>api/serial</code> endpoint</li>
<li><code>event_listener</code> for <strong>post</strong> request to <code>api/skills/event</code> endpoint</li>
<li><code>website_show</code> for <strong>post</strong> request to <code>api/webviews/display</code> endpoint</li>
<li><code>website_settings</code> for <strong>post</strong> request to <code>api/webviews/settings</code> endpoint</li>
<li><code>blink_on</code> for <strong>post</strong> request to <code>api/blink</code> endpoint</li>
<li><code>blink_settings</code> for <strong>post</strong> request to <code>api/blink/settings</code> endpoint</li>
<li><code>display_settings</code> for <strong>post</strong> request to <code>api/display/settings</code> endpoint</li>
<li><code>flashlight_on</code> for <strong>post</strong> request to <code>api/flashlight</code> endpoint</li>
<li><code>speak</code> for <strong>post</strong> request to <code>api/tts/speak</code> endpoint</li>
<li><code>speak_stop</code> for <strong>post</strong> request to <code>api/tts/stop</code> endpoint</li>
<li><code>speech_capture</code> for <strong>post</strong> request to <code>api/audio/speech/capture</code> endpoint</li>
<li><code>drive</code> for <strong>post</strong> request to <code>api/drive</code> endpoint</li>
<li><code>drive_arc</code> for <strong>post</strong> request to <code>api/drive/arc</code> endpoint</li>
<li><code>drive_heading</code> for <strong>post</strong> request to <code>api/drive/hdt</code> endpoint</li>
<li><code>drive_time</code> for <strong>post</strong> request to <code>api/drive/time</code> endpoint</li>
<li><code>drive_track</code> for <strong>post</strong> request to <code>api/drive/track</code> endpoint</li>
<li><code>drive_stop</code> for <strong>post</strong> request to <code>api/drive/stop</code> endpoint</li>
<li><code>drive_to_loc</code> for <strong>post</strong> request to <code>api/drive/coordinates</code> endpoint</li>
<li><code>drive_on_path</code> for <strong>post</strong> request to <code>api/drive/path</code> endpoint</li>
<li><code>halt</code> for <strong>post</strong> request to <code>api/halt</code> endpoint</li>
<li><code>arm_move</code> for <strong>post</strong> request to <code>api/arms</code> endpoint</li>
<li><code>arms_move</code> for <strong>post</strong> request to <code>api/arms/set</code> endpoint</li>
<li><code>head_move</code> for <strong>post</strong> request to <code>api/head</code> endpoint</li>
<li><code>hazard_settings</code> for <strong>post</strong> request to <code>api/hazard/updatebasesettings</code> endpoint</li>
<li><code>streaming_av_start</code> for <strong>post</strong> request to <code>api/avstreaming/start</code> endpoint</li>
<li><code>streaming_av_stop</code> for <strong>post</strong> request to <code>api/avstreaming/stop</code> endpoint</li>
<li><code>streaming_av_disable</code> for <strong>post</strong> request to <code>api/services/avstreaming/disable</code> endpoint</li>
<li><code>streaming_av_enable</code> for <strong>post</strong> request to <code>api/services/avstreaming/enable</code> endpoint</li>
<li><code>keyphrase_recognition_start</code> for <strong>post</strong> request to <code>api/audio/keyphrase/start</code> endpoint</li>
<li><code>keyphrase_recognition_stop</code> for <strong>post</strong> request to <code>api/audio/keyphrase/stop</code> endpoint</li>
<li><code>update_allow</code> for <strong>post</strong> request to <code>api/system/update/allow</code> endpoint</li>
<li><code>update_perform</code> for <strong>post</strong> request to <code>api/system/update</code> endpoint</li>
<li><code>update_perform_targeted</code> for <strong>post</strong> request to <code>api/system/update/component</code> endpoint</li>
<li><code>update_prevent</code> for <strong>post</strong> request to <code>api/system/update/prevent</code> endpoint</li>
<li><code>error_text_clear</code> for <strong>post</strong> request to <code>api/text/error/clear</code> endpoint</li>
<li><code>camera_disable</code> for <strong>post</strong> request to <code>api/services/camera/disable</code> endpoint</li>
<li><code>camera_enable</code> for <strong>post</strong> request to <code>api/services/camera/enable</code> endpoint</li>
<li><code>restart</code> for <strong>post</strong> request to <code>api/reboot</code> endpoint</li>
<li><code>volume_settings</code> for <strong>post</strong> request to <code>api/audio/volume</code> endpoint</li>
<li><code>logs_settings</code> for <strong>post</strong> request to <code>api/logs/level</code> endpoint</li>
<li><code>websocket_settings</code> for <strong>post</strong> request to <code>api/websocket/version</code> endpoint</li>
<li><code>external_request</code> for <strong>post</strong> request to <code>api/request</code> endpoint</li>
</ul>
</details>

<details> 
  <summary>List of supported information keywords</summary>

<ul>
<li><code>audio_file</code> for <strong>get</strong> request to <code>api/audio</code> endpoint</li>
<li><code>audio_list</code> for <strong>get</strong> request to <code>api/audio/list</code> endpoint</li>
<li><code>audio_status</code> for <strong>get</strong> request to <code>api/services/audio</code> endpoint</li>
<li><code>image_file</code> for <strong>get</strong> request to <code>api/images</code> endpoint</li>
<li><code>image_list</code> for <strong>get</strong> request to <code>api/images/list</code> endpoint</li>
<li><code>video_file</code> for <strong>get</strong> request to <code>api/videos</code> endpoint</li>
<li><code>video_list</code> for <strong>get</strong> request to <code>api/videos/list</code> endpoint</li>
<li><code>av_status</code> for <strong>get</strong> request to <code>api/services/avstreaming</code> endpoint</li>
<li><code>sensor_values</code> for <strong>get</strong> request to <code>api/serial</code> endpoint</li>
<li><code>map_file</code> for <strong>get</strong> request to <code>api/slam/map</code> endpoint</li>
<li><code>current_map_id</code> for <strong>get</strong> request to <code>api/slam/map/current</code> endpoint</li>
<li><code>map_id_list</code> for <strong>get</strong> request to <code>api/slam/map/ids</code> endpoint</li>
<li><code>slam_diagnostics</code> for <strong>get</strong> request to <code>api/slam/diagnostics</code> endpoint</li>
<li><code>slam_path</code> for <strong>get</strong> request to <code>api/slam/path</code> endpoint</li>
<li><code>slam_status</code> for <strong>get</strong> request to <code>api/slam/status</code> endpoint</li>
<li><code>slam_enabled</code> for <strong>get</strong> request to <code>api/services/slam</code> endpoint</li>
<li><code>picture_depth</code> for <strong>get</strong> request to <code>api/cameras/depth</code> endpoint</li>
<li><code>picture_fisheye</code> for <strong>get</strong> request to <code>api/cameras/fisheye</code> endpoint</li>
<li><code>picture_rgb</code> for <strong>get</strong> request to <code>api/cameras/rgb</code> endpoint</li>
<li><code>faces_known</code> for <strong>get</strong> request to <code>api/faces</code> endpoint</li>
<li><code>recording_file</code> for <strong>get</strong> request to <code>api/videos/recordings</code> endpoint</li>
<li><code>recording_list</code> for <strong>get</strong> request to <code>api/videos/recordings/list</code> endpoint</li>
<li><code>skills_running</code> for <strong>get</strong> request to <code>api/skills/running</code> endpoint</li>
<li><code>skills_known</code> for <strong>get</strong> request to <code>api/skills</code> endpoint</li>
<li><code>wifis_available</code> for <strong>get</strong> request to <code>api/networks/scan</code> endpoint</li>
<li><code>wifis_saved</code> for <strong>get</strong> request to <code>api/networks</code> endpoint</li>
<li><code>battery_status</code> for <strong>get</strong> request to <code>api/battery</code> endpoint</li>
<li><code>camera_status</code> for <strong>get</strong> request to <code>api/services/camera</code> endpoint</li>
<li><code>blink_settings</code> for <strong>get</strong> request to <code>api/blink/settings</code> endpoint</li>
<li><code>hazards_settings</code> for <strong>get</strong> request to <code>api/hazards/settings</code> endpoint</li>
<li><code>camera_settings</code> for <strong>get</strong> request to <code>api/camera</code> endpoint</li>
<li><code>slam_visible_settings</code> for <strong>get</strong> request to <code>api/slam/settings/visible</code> endpoint</li>
<li><code>slam_infrared_settings</code> for <strong>get</strong> request to <code>api/slam/settings/ir</code> endpoint</li>
<li><code>update_settings</code> for <strong>get</strong> request to <code>api/system/update/settings</code> endpoint</li>
<li><code>device</code> for <strong>get</strong> request to <code>api/device</code> endpoint</li>
<li><code>help</code> for <strong>get</strong> request to <code>api/help</code> endpoint</li>
<li><code>log</code> for <strong>get</strong> request to <code>api/logs</code> endpoint</li>
<li><code>log_level</code> for <strong>get</strong> request to <code>api/logs/level</code> endpoint</li>
<li><code>update_available</code> for <strong>get</strong> request to <code>api/system/updates</code> endpoint</li>
<li><code>websockets</code> for <strong>get</strong> request to <code>api/websockets</code> endpoint</li>
<li><code>websocket_version</code> for <strong>get</strong> request to <code>api/websocket/version</code></li>
</ul>
</details>

<details> 
  <summary>List of supported data shortcuts</summary>

<ul>
<li><code>led_off</code> for <code>{&quot;red&quot;: &quot;0&quot;, &quot;green&quot;: &quot;0&quot;, &quot;blue&quot;: &quot;0&quot;}</code></li>
<li><code>white_light</code> for <code>{&quot;red&quot;: &quot;255&quot;, &quot;green&quot;: &quot;255&quot;, &quot;blue&quot;: &quot;255&quot;}</code></li>
<li><code>red_light</code> for <code>{&quot;red&quot;: &quot;255&quot;, &quot;green&quot;: &quot;0&quot;, &quot;blue&quot;: &quot;0&quot;}</code></li>
<li><code>green_light</code> for <code>{&quot;red&quot;: &quot;0&quot;, &quot;green&quot;: &quot;255&quot;, &quot;blue&quot;: &quot;0&quot;}</code></li>
<li><code>blue_light</code> for <code>{&quot;red&quot;: &quot;0&quot;, &quot;green&quot;: &quot;0&quot;, &quot;blue&quot;: &quot;255&quot;}</code></li>
<li><code>yellow_light</code> for <code>{&quot;red&quot;: &quot;255&quot;, &quot;green&quot;: &quot;255&quot;, &quot;blue&quot;: &quot;0&quot;}</code></li>
<li><code>cyan_light</code> for <code>{&quot;red&quot;: &quot;0&quot;, &quot;green&quot;: &quot;255&quot;, &quot;blue&quot;: &quot;255&quot;}</code></li>
<li><code>magenta_light</code> for <code>{&quot;red&quot;: &quot;255&quot;, &quot;green&quot;: &quot;0&quot;, &quot;blue&quot;: &quot;255&quot;}</code></li>
<li><code>orange_light</code> for <code>{&quot;red&quot;: &quot;255&quot;, &quot;green&quot;: &quot;125&quot;, &quot;blue&quot;: &quot;0&quot;}</code></li>
<li><code>lime_light</code> for <code>{&quot;red&quot;: &quot;125&quot;, &quot;green&quot;: &quot;255&quot;, &quot;blue&quot;: &quot;0&quot;}</code></li>
<li><code>aqua_light</code> for <code>{&quot;red&quot;: &quot;0&quot;, &quot;green&quot;: &quot;255&quot;, &quot;blue&quot;: &quot;125&quot;}</code></li>
<li><code>azure_light</code> for <code>{&quot;red&quot;: &quot;0&quot;, &quot;green&quot;: &quot;125&quot;, &quot;blue&quot;: &quot;255&quot;}</code></li>
<li><code>violet_light</code> for <code>{&quot;red&quot;: &quot;125&quot;, &quot;green&quot;: &quot;0&quot;, &quot;blue&quot;: &quot;255&quot;}</code></li>
<li><code>pink_light</code> for <code>{&quot;red&quot;: &quot;255&quot;, &quot;green&quot;: &quot;0&quot;, &quot;blue&quot;: &quot;125&quot;}</code></li>
<li><code>low_volume</code> for <code>{&quot;Volume&quot;: &quot;5&quot;}</code></li>
<li><code>image_admiration</code> for <code>{&quot;FileName&quot;: &quot;e_Admiration.jpg&quot;}</code></li>
<li><code>image_aggressiveness</code> for <code>{&quot;FileName&quot;: &quot;e_Aggressiveness.jpg&quot;}</code></li>
<li><code>image_amazement</code> for <code>{&quot;FileName&quot;: &quot;e_Amazement.jpg&quot;}</code></li>
<li><code>image_anger</code> for <code>{&quot;FileName&quot;: &quot;e_Anger.jpg&quot;}</code></li>
<li><code>image_concerned</code> for <code>{&quot;FileName&quot;: &quot;e_ApprehensionConcerned.jpg&quot;}</code></li>
<li><code>image_contempt</code> for <code>{&quot;FileName&quot;: &quot;e_Contempt.jpg&quot;}</code></li>
<li><code>image_content_left</code> for <code>{&quot;FileName&quot;: &quot;e_ContentLeft.jpg&quot;}</code></li>
<li><code>image_content_right</code> for <code>{&quot;FileName&quot;: &quot;e_ContentRight.jpg&quot;}</code></li>
<li><code>image_content_default</code> for <code>{&quot;FileName&quot;: &quot;e_DefaultContent.jpg&quot;}</code></li>
<li><code>image_disgust</code> for <code>{&quot;FileName&quot;: &quot;e_Disgust.jpg&quot;}</code></li>
<li><code>image_disoriented</code> for <code>{&quot;FileName&quot;: &quot;e_Disoriented.jpg&quot;}</code></li>
<li><code>image_hilarious</code> for <code>{&quot;FileName&quot;: &quot;e_EcstacyHilarious.jpg&quot;}</code></li>
<li><code>image_starry_eyed</code> for <code>{&quot;FileName&quot;: &quot;e_EcstacyStarryEyed.jpg&quot;}</code></li>
<li><code>image_fear</code> for <code>{&quot;FileName&quot;: &quot;e_Fear.jpg&quot;}</code></li>
<li><code>image_grief</code> for <code>{&quot;FileName&quot;: &quot;e_Grief.jpg&quot;}</code></li>
<li><code>image_joy_1</code> for <code>{&quot;FileName&quot;: &quot;e_Joy.jpg&quot;}</code></li>
<li><code>image_joy_2</code> for <code>{&quot;FileName&quot;: &quot;e_Joy2.jpg&quot;}</code></li>
<li><code>image_goofy_1</code> for <code>{&quot;FileName&quot;: &quot;e_JoyGoofy.jpg&quot;}</code></li>
<li><code>image_goofy_2</code> for <code>{&quot;FileName&quot;: &quot;e_JoyGoofy2.jpg&quot;}</code></li>
<li><code>image_goofy_3</code> for <code>{&quot;FileName&quot;: &quot;e_JoyGoofy3.jpg&quot;}</code></li>
<li><code>image_love</code> for <code>{&quot;FileName&quot;: &quot;e_Love.jpg&quot;}</code></li>
<li><code>image_rage_1</code> for <code>{&quot;FileName&quot;: &quot;e_Rage.jpg&quot;}</code></li>
<li><code>image_rage_2</code> for <code>{&quot;FileName&quot;: &quot;e_Rage2.jpg&quot;}</code></li>
<li><code>image_rage_3</code> for <code>{&quot;FileName&quot;: &quot;e_Rage3.jpg&quot;}</code></li>
<li><code>image_rage_4</code> for <code>{&quot;FileName&quot;: &quot;e_Rage4.jpg&quot;}</code></li>
<li><code>image_remorse</code> for <code>{&quot;FileName&quot;: &quot;e_RemorseShame.jpg&quot;}</code></li>
<li><code>image_sadness</code> for <code>{&quot;FileName&quot;: &quot;e_Sadness.jpg&quot;}</code></li>
<li><code>image_sleping_1</code> for <code>{&quot;FileName&quot;: &quot;e_Sleeping.jpg&quot;}</code></li>
<li><code>image_sleeping_2</code> for <code>{&quot;FileName&quot;: &quot;e_SleepingZZZ.jpg&quot;}</code></li>
<li><code>image_sleepy_1</code> for <code>{&quot;FileName&quot;: &quot;e_Sleepy.jpg&quot;}</code></li>
<li><code>image_sleepy_2</code> for <code>{&quot;FileName&quot;: &quot;e_Sleepy2.jpg&quot;}</code></li>
<li><code>image_sleepy_3</code> for <code>{&quot;FileName&quot;: &quot;e_Sleepy3.jpg&quot;}</code></li>
<li><code>image_sleepy_4</code> for <code>{&quot;FileName&quot;: &quot;e_Sleepy4.jpg&quot;}</code></li>
<li><code>image_surprise</code> for <code>{&quot;FileName&quot;: &quot;e_Surprise.jpg&quot;}</code></li>
<li><code>image_system_black_screen</code> for <code>{&quot;FileName&quot;: &quot;e_SystemBlackScreen.jpg&quot;}</code></li>
<li><code>image_system_blink_large</code> for <code>{&quot;FileName&quot;: &quot;e_SystemBlinkLarge.jpg&quot;}</code></li>
<li><code>image_system_blink_standard</code> for <code>{&quot;FileName&quot;: &quot;e_SystemBlinkStandard.jpg&quot;}</code></li>
<li><code>image_system_camera</code> for <code>{&quot;FileName&quot;: &quot;e_SystemCamera.jpg&quot;}</code></li>
<li><code>image_system_flash</code> for <code>{&quot;FileName&quot;: &quot;e_SystemFlash.jpg&quot;}</code></li>
<li><code>image_system_gear_prompt</code> for <code>{&quot;FileName&quot;: &quot;e_SystemGearPrompt.jpg&quot;}</code></li>
<li><code>image_system_logo_prompt</code> for <code>{&quot;FileName&quot;: &quot;e_SystemLogoPrompt.jpg&quot;}</code></li>
<li><code>image_terror_1</code> for <code>{&quot;FileName&quot;: &quot;e_Terror.jpg&quot;}</code></li>
<li><code>image_terror_2</code> for <code>{&quot;FileName&quot;: &quot;e_Terror2.jpg&quot;}</code></li>
<li><code>image_terror_left</code> for <code>{&quot;FileName&quot;: &quot;e_TerrorLeft.jpg&quot;}</code></li>
<li><code>image_terror_right</code> for <code>{&quot;FileName&quot;: &quot;e_TerrorRight.jpg&quot;}</code></li>
<li><code>sound_acceptance</code> for <code>{&quot;FileName&quot;: &quot;s_Acceptance.wav&quot;}</code></li>
<li><code>sound_amazement_1</code> for <code>{&quot;FileName&quot;: &quot;s_Amazement.wav&quot;}</code></li>
<li><code>sound_amazement_2</code> for <code>{&quot;FileName&quot;: &quot;s_Amazement2.wav&quot;}</code></li>
<li><code>sound_anger_1</code> for <code>{&quot;FileName&quot;: &quot;s_Anger.wav&quot;}</code></li>
<li><code>sound_anger_2</code> for <code>{&quot;FileName&quot;: &quot;s_Anger2.wav&quot;}</code></li>
<li><code>sound_anger_3</code> for <code>{&quot;FileName&quot;: &quot;s_Anger3.wav&quot;}</code></li>
<li><code>sound_anger_4</code> for <code>{&quot;FileName&quot;: &quot;s_Anger4.wav&quot;}</code></li>
<li><code>sound_annoyance_1</code> for <code>{&quot;FileName&quot;: &quot;s_Annoyance.wav&quot;}</code></li>
<li><code>sound_annoyance_2</code> for <code>{&quot;FileName&quot;: &quot;s_Annoyance2.wav&quot;}</code></li>
<li><code>sound_annoyance_3</code> for <code>{&quot;FileName&quot;: &quot;s_Annoyance3.wav&quot;}</code></li>
<li><code>sound_annoyance_4</code> for <code>{&quot;FileName&quot;: &quot;s_Annoyance4.wav&quot;}</code></li>
<li><code>sound_awe_1</code> for <code>{&quot;FileName&quot;: &quot;s_Awe.wav&quot;}</code></li>
<li><code>sound_awe_2</code> for <code>{&quot;FileName&quot;: &quot;s_Awe2.wav&quot;}</code></li>
<li><code>sound_awe_3</code> for <code>{&quot;FileName&quot;: &quot;s_Awe3.wav&quot;}</code></li>
<li><code>sound_boredom</code> for <code>{&quot;FileName&quot;: &quot;s_Boredom.wav&quot;}</code></li>
<li><code>sound_disapproval</code> for <code>{&quot;FileName&quot;: &quot;s_Disapproval.wav&quot;}</code></li>
<li><code>sound_disgust_1</code> for <code>{&quot;FileName&quot;: &quot;s_Disgust.wav&quot;}</code></li>
<li><code>sound_disgust_2</code> for <code>{&quot;FileName&quot;: &quot;s_Disgust2.wav&quot;}</code></li>
<li><code>sound_disgust_3</code> for <code>{&quot;FileName&quot;: &quot;s_Disgust3.wav&quot;}</code></li>
<li><code>sound_disoriented_1</code> for <code>{&quot;FileName&quot;: &quot;s_DisorientedConfused.wav&quot;}</code></li>
<li><code>sound_disoriented_2</code> for <code>{&quot;FileName&quot;: &quot;s_DisorientedConfused2.wav&quot;}</code></li>
<li><code>sound_disoriented_3</code> for <code>{&quot;FileName&quot;: &quot;s_DisorientedConfused3.wav&quot;}</code></li>
<li><code>sound_disoriented_4</code> for <code>{&quot;FileName&quot;: &quot;s_DisorientedConfused4.wav&quot;}</code></li>
<li><code>sound_disoriented_5</code> for <code>{&quot;FileName&quot;: &quot;s_DisorientedConfused5.wav&quot;}</code></li>
<li><code>sound_disoriented_6</code> for <code>{&quot;FileName&quot;: &quot;s_DisorientedConfused6.wav&quot;}</code></li>
<li><code>sound_distraction</code> for <code>{&quot;FileName&quot;: &quot;s_Distraction.wav&quot;}</code></li>
<li><code>sound_ecstacy_1</code> for <code>{&quot;FileName&quot;: &quot;s_Ecstacy.wav&quot;}</code></li>
<li><code>sound_ecstacy_2</code> for <code>{&quot;FileName&quot;: &quot;s_Ecstacy2.wav&quot;}</code></li>
<li><code>sound_fear</code> for <code>{&quot;FileName&quot;: &quot;s_Fear.wav&quot;}</code></li>
<li><code>sound_grief_1</code> for <code>{&quot;FileName&quot;: &quot;s_Grief.wav&quot;}</code></li>
<li><code>sound_grief_2</code> for <code>{&quot;FileName&quot;: &quot;s_Grief2.wav&quot;}</code></li>
<li><code>sound_grief_3</code> for <code>{&quot;FileName&quot;: &quot;s_Grief3.wav&quot;}</code></li>
<li><code>sound_grief_4</code> for <code>{&quot;FileName&quot;: &quot;s_Grief4.wav&quot;}</code></li>
<li><code>sound_joy_1</code> for <code>{&quot;FileName&quot;: &quot;s_Joy.wav&quot;}</code></li>
<li><code>sound_joy_2</code> for <code>{&quot;FileName&quot;: &quot;s_Joy2.wav&quot;}</code></li>
<li><code>sound_joy_3</code> for <code>{&quot;FileName&quot;: &quot;s_Joy3.wav&quot;}</code></li>
<li><code>sound_joy_4</code> for <code>{&quot;FileName&quot;: &quot;s_Joy4.wav&quot;}</code></li>
<li><code>sound_loathing</code> for <code>{&quot;FileName&quot;: &quot;s_Loathing.wav&quot;}</code></li>
<li><code>sound_love</code> for <code>{&quot;FileName&quot;: &quot;s_Love.wav&quot;}</code></li>
<li><code>sound_phrase_bye_bye</code> for <code>{&quot;FileName&quot;: &quot;s_PhraseByeBye.wav&quot;}</code></li>
<li><code>sound_phrase_evil</code> for <code>{&quot;FileName&quot;: &quot;s_PhraseEvilAhHa.wav&quot;}</code></li>
<li><code>sound_phrase_hello</code> for <code>{&quot;FileName&quot;: &quot;s_PhraseHello.wav&quot;}</code></li>
<li><code>sound_phrase_no</code> for <code>{&quot;FileName&quot;: &quot;s_PhraseNoNoNo.wav&quot;}</code></li>
<li><code>sound_phrase_oopsy</code> for <code>{&quot;FileName&quot;: &quot;s_PhraseOopsy.wav&quot;}</code></li>
<li><code>sound_phrase_ow</code> for <code>{&quot;FileName&quot;: &quot;s_PhraseOwOwOw.wav&quot;}</code></li>
<li><code>sound_phrase_oww</code> for <code>{&quot;FileName&quot;: &quot;s_PhraseOwwww.wav&quot;}</code></li>
<li><code>sound_phrase_uh</code> for <code>{&quot;FileName&quot;: &quot;s_PhraseUhOh.wav&quot;}</code></li>
<li><code>sound_rage</code> for <code>{&quot;FileName&quot;: &quot;s_Rage.wav&quot;}</code></li>
<li><code>sound_sadness_1</code> for <code>{&quot;FileName&quot;: &quot;s_Sadness.wav&quot;}</code></li>
<li><code>sound_sadness_2</code> for <code>{&quot;FileName&quot;: &quot;s_Sadness2.wav&quot;}</code></li>
<li><code>sound_sadness_3</code> for <code>{&quot;FileName&quot;: &quot;s_Sadness3.wav&quot;}</code></li>
<li><code>sound_sadness_4</code> for <code>{&quot;FileName&quot;: &quot;s_Sadness4.wav&quot;}</code></li>
<li><code>sound_sadness_5</code> for <code>{&quot;FileName&quot;: &quot;s_Sadness5.wav&quot;}</code></li>
<li><code>sound_sadness_6</code> for <code>{&quot;FileName&quot;: &quot;s_Sadness6.wav&quot;}</code></li>
<li><code>sound_sadness_7</code> for <code>{&quot;FileName&quot;: &quot;s_Sadness7.wav&quot;}</code></li>
<li><code>sound_sleepy_1</code> for <code>{&quot;FileName&quot;: &quot;s_Sleepy.wav&quot;}</code></li>
<li><code>sound_sleepy_2</code> for <code>{&quot;FileName&quot;: &quot;s_Sleepy2.wav&quot;}</code></li>
<li><code>sound_sleepy_3</code> for <code>{&quot;FileName&quot;: &quot;s_Sleepy3.wav&quot;}</code></li>
<li><code>sound_sleepy_4</code> for <code>{&quot;FileName&quot;: &quot;s_Sleepy4.wav&quot;}</code></li>
<li><code>sound_sleepy_snore</code> for <code>{&quot;FileName&quot;: &quot;s_SleepySnore.wav&quot;}</code></li>
<li><code>sound_camera_shutter</code> for <code>{&quot;FileName&quot;: &quot;s_SystemCameraShutter.wav&quot;}</code></li>
<li><code>sound_failure</code> for <code>{&quot;FileName&quot;: &quot;s_SystemFailure.wav&quot;}</code></li>
<li><code>sound_success</code> for <code>{&quot;FileName&quot;: &quot;s_SystemSuccess.wav&quot;}</code></li>
<li><code>sound_wake</code> for <code>{&quot;FileName&quot;: &quot;s_SystemWakeWord.wav&quot;}</code></li>
</ul>
</details>
"""



"""The module contains the class Misty which represents a Misty II robot.
"""
from misty2py.action import *
from misty2py.information import *
from misty2py.misty_event import MistyEventHandler


class Misty:
    """A class representing a Misty robot.

    Attributes:
        ip (str): The IP address of this Misty robot.
        infos (Info): The object of Info class that belongs to this Misty.
        actions (Action): The object of Action class that belongs to this Misty.
        events (dict): A dictionary of active event subscriptions (keys being the event name, values the MistyEvent() object).
    """

    def __init__(
        self,
        ip: str,
        custom_info: Dict = {},
        custom_actions: Dict = {},
        custom_data: Dict = {},
    ):
        """Initialises an instance of a Misty robot.

        Args:
            ip (str): The IP address of this Misty robot.
            custom_info (Dict, optional): Custom information keywords in the form of dictionary with key being the keyword and value being the API endpoint. Defaults to `{}`.
            custom_actions (Dict, optional): Custom actions keywords in the form of dictionary with key being the keyword and value being the API endpoint. Defaults to `{}`.
            custom_data (Dict, optional): Custom data shortcuts in the form of dictionary with key being the shortcut and value being the json data in the form of a dictionary. Defaults to `{}`.
        """

        self.ip = ip
        self.infos = Info(ip, custom_allowed_infos=custom_info)
        self.actions = Action(
            ip,
            custom_allowed_actions=custom_actions,
            custom_allowed_data=custom_data
        )
        self.event_handler = MistyEventHandler(ip)

    def __str__(self) -> str:
        """Parses a Misty object into a string.

        Returns:
            str: A string identifiyng this Misty object.
        """

        return "A Misty II robot with IP address %s" % self.ip

    def perform_action(self, action_name: str, data: Dict = {}) -> Dict:
        """Sends Misty a request to perform an action.

        Args:
            action_name (str): The keyword specifying the action to perform.
            data (Dict, optional): The data to send in the request body in the form of a data shortcut or a json dictionary. Defaults to `{}`.

        Returns:
            Dict: The response from Misty's REST API.
        """

        return self.actions.action_handler(action_name, data)

    def get_info(self, info_name: str, params: Dict = {}) -> Dict:
        """Obtains information from Misty.

        Args:
            info_name (str): The information keyword specifying which kind of information to retrieve.
            params (Dict): A dictionary of parameter names and parameter values. Defaults to `{}`.

        Returns:
            Dict: The response from Misty's REST API.
        """

        return self.infos.get_info(info_name, params)

    def event(self, action: str, **kwargs) -> Dict:
        """Handles event-related actions.

        Supports following actions:

        - **event subscripton** - requires an action keyword `subscribe` and an argument `type` (a string representing the event type to subscribe to). Optional arguments are:
            - `name` (str) for a custom event name; must be unique.
            - `return_property` (str) for the property to return from Misty's WebSockets API; all properties are returned if `return_property` is not supplied.
            - debounce (int) for the interval at which new information is sent in ms; defaults to `250`.
            - len_data_entries (int) for the maximum number of data entries to keep (discards in fifo style); defaults to `10`.
            - event_emitter (Callable) for an event emitter function which emits an event upon message recieval; defaults to `None`.
        - **obtaining the data from an event** - requires an action keyword `get_data` and an argument `name` (the event name).
        - **obtaining the log from an event** - requires an action keyword `get_log` and an argument `name` (the event name).
        - **unsubscribing from an event** - requires an action keyword `unsubscribe` and an argument `name` (the event name).

        Args:
            action (str): The action keyword.

        Returns:
            Dict: The JSON dictionary response.
        """

        if action == "subscribe":
            return self.event_handler.subscribe_event(kwargs)

        if action == "get_data":
            return self.event_handler.get_event_data(kwargs)

        if action == "get_log":
            return self.event_handler.get_event_log(kwargs)

        if action == "unsubscribe":
            return self.event_handler.unsubscribe_event(kwargs)

        return {
            "status": "Failed",
            "message": "Unknown event action: `%s`." % action
        }



"""This module contains the class Status to track the execution status of a script and the class ActionLog to track the actions performed in a script."""
from typing import Any, Dict, List


class Status:
    """Tracks the status of a behaviour."""

    def __init__(
        self,
        init_status: Any = "initialised",
        init_data: Any = "",
        init_time: float = 0,
    ) -> None:
        """Initialises the Status.

        Args:
            init_status (Any, optional): The initial status. Defaults to `"initialised"`.
            init_data (Any, optional): The initial data. Defaults to `""`.
            init_time (float, optional): The initial time. Defaults to `0`.
        """
        self.status = init_status
        self.data = init_data
        self.time = init_time

    def set_(self, **content) -> None:
        """Sets the parameter passed to the function to the value passed.

        Accepts parameters `"data"`, `"time"` and `"status"`.
        """
        potential_data = content.get("data")
        if not isinstance(potential_data, type(None)):
            self.data = potential_data

        potential_time = content.get("time")
        if not isinstance(potential_time, type(None)):
            self.time = potential_time

        potential_status = content.get("status")
        if not isinstance(potential_status, type(None)):
            self.status = potential_status

    def get_(self, content_type: str) -> Any:
        """Obtains the value of the specified parameter.

        Args:
            content_type (str): The parameter whose value to return. Accepts `"data"`, `"time"` and `"status"`.

        Returns:
            Any: The value of the requested parameter.
        """
        if content_type == "data":
            return self.data
        if content_type == "time":
            return self.time
        if content_type == "status":
            return self.status

    def parse_to_message(self) -> Dict:
        """Returns the current status as a JSON dictionary.

        Returns:
            Dict: The current status, always contains the keyword `"status"`, optionally contains the keywords `"time"` and `"data"`.
        """
        message = {}
        if isinstance(self.status, bool):
            if self.status:
                message["status"] = "Success"
            else:
                message["status"] = "Failed"
        else:
            message["status"] = self.status
        if self.time != 0:
            message["time"] = self.time
        if self.data != "":
            message["data"] = self.data
        return message


class ActionLog:
    """Logs performed actions."""

    def __init__(self) -> None:
        """Initialises the actions list."""
        self.actions = []

    def append_(self, value: Any):
        """Appends the value to the actions list."""
        self.actions.append(value)

    def get_(self) -> List:
        """Returns the actions list."""
        return self.actions



"""This module contains base64 encoders and decoders."""
import base64
from typing import Union


def content_to_base64(
    input_str: str, is_path: bool = False, save_path: Union[str, bool] = False
) -> str:
    """Encodes content to base64.

    Args:
        input_str (str): The utf-8 string to encode or the path to the file to encode.
        is_path (bool, optional): `True` if `input_str` is the path to the file to encode or `False` if `input_str` is the content to encode. Defaults to `False`.
        save_path (Union[str, bool], optional): The path to the file where the encoded content should be saved or `False` if the encoded content should be returned as utf-8 string. Defaults to `False`.

    Returns:
        str: The encoded content if `save_path` is `False` or the success message if `save_path` is `True`.
    """
    if is_path:
        try:
            data = open(input_str, "rb").read()
        except Exception as e:
            return "Error reading the file `%s`. Error message: `%s`" % (
                input_str,
                e
            )
    else:
        data = input_str.encode()
    encoded = base64.b64encode(data)
    if save_path:
        try:
            with open(save_path, "wb") as f:
                f.write(encoded)
            return "Successfully saved to `%s`" % save_path
        except Exception as e:
            return "Error saving to `%s`. Error message: `%s`" % (
                save_path,
                e
            )
    return encoded.decode("utf-8")


def base64_to_content(
    input_str: str, is_path: bool = False, save_path: Union[str, bool] = False
) -> str:
    """Decodes base64 to utf-8 string or other content type specified by the extension in `save_path`.

    Args:
        input_str (str): The utf-8 string to decode or the path to the file to decode.
        is_path (bool, optional): `True` if `input_str` is the path to the file to decode, `False` if `input_str` is the content to decode. Defaults to `False`.
        save_path (Union[str,bool], optional): The path to the file where the decoded content should be saved or `False` if the decoded content should be returned as utf-8 string. Defaults to `False`.

    Returns:
        str: The decoded content or the success string if `save_path`.
    """
    if is_path:
        try:
            data = open(input_str, "rb").read()
        except Exception as e:
            return "Error reading the file `%s`. Error message: `%s`" % (
                input_str,
                e
            )
    else:
        data = input_str.encode()
    decoded = base64.b64decode(data)
    if save_path:
        try:
            with open(save_path, "wb") as f:
                f.write(decoded)
            return "Successfully saved to `%s`" % save_path
        except Exception as e:
            return "Error saving to `%s`. Error message: `%s`" % (
                save_path,
                e
            )
    return decoded.decode("utf-8")



"""This module's function is to send information requests via the information keywords matching to Misty's API endpoints, sending information requests and matching data shortcuts.
"""
import json
from os import path
from typing import Dict
from urllib.parse import urlencode

import requests

this_directory = path.abspath(path.dirname(__file__))
INFOS_JSON = str(path.join(this_directory, "allowed_infos.json"))


class Get:
    """A class representing the GET url request method.

    Attributes:
        ip (str): The IP address where the requests are sent
        allowed_infos (dict): The dictionary of information keywords matching to the Misty's API endpoints.
    """

    def __init__(self, ip: str, custom_allowed_infos: Dict = {}) -> None:
        """Initialises a Get object.

        Args:
            ip (str): The IP address where the requests are sent.
            custom_allowed_infos (Dict, optional): The dictionary of custom information keywords. Defaults to `{}`.
        """

        self.ip = ip

        allowed_infos = custom_allowed_infos
        f = open(INFOS_JSON)
        allowed_infos.update(json.loads(f.read()))
        f.close()

        self.allowed_infos = allowed_infos

    def get_info(self, endpoint: str) -> Dict:
        """Sends a GET request.

        Args:
            endpoint (str): The API endpoint to which the request is sent.

        Returns:
            Dict: The response from Misty's REST API.
        """

        r = requests.get("http://%s/%s" % (self.ip, endpoint))
        try:
            return r.json()
        except Exception as e:
            return {"status": "Success", "content": r.content, "error_msg": e}


class Info(Get):
    """A class representing an information request from Misty.
    A subclass of Get()."""

    def get_info(self, info_name: str, params: Dict = {}) -> Dict:
        """Sends an information request to Misty.

        Args:
            info_name (str): The information keyword specifying which information is requested.
            params (Dict): dict of parameter name and parameter value. Defaults to `{}`.

        Returns:
            Dict: The response from Misty's REST API.
        """

        if info_name not in self.allowed_infos.keys():
            response = {
                "status": "Failed",
                "message": "Command `%s` not supported." % info_name,
            }

        else:
            endpoint = self.allowed_infos[info_name]

            if len(params) > 0:
                endpoint += "?"
                query_string = urlencode(params)
                endpoint += query_string

            try:
                response = super().get_info(endpoint)

            except Exception as e:
                response = {
                    "status": "Failed",
                    "message": "Error occured. Error message: `%s`" % e,
                }

        return response



"""This module handles Misty's movement."""
from typing import Dict

from misty2py.robot import Misty


class Movement:
    """Handles Misty's movement."""

    def __init__(
        self,
        max_velocity=100,
        min_velocity=1,
        stagnant_velocity=0,
        max_angle=100,
        min_angle=1,
        stagnant_angle=0,
    ) -> None:
        """Initialises the Movement class.

        Args:
            max_velocity (int, optional): The maximal allowed velocity. Defaults to `100`.
            min_velocity (int, optional): The minimal allowed velocity that is not stagnant. Defaults to `1`.
            stagnant_velocity (int, optional): The stagnant velocity. Defaults to `0`.
            max_angle (int, optional): The maximum turning angle. Defaults to `100`.
            min_angle (int, optional): The minimum turning angle that is not equal to the straight line driving. Defaults to `1`.
            stagnant_angle (int, optional): The angle equal to the straight line driving. Defaults to `0`.
        """
        self.max_velocity = max_velocity
        self.min_velocity = min_velocity
        self.stagnant_velocity = stagnant_velocity
        self.max_angle = max_angle
        self.min_angle = min_angle
        self.stagnant_angle = stagnant_angle

    def _parse_value(
        self,
        value: int,
        max: int,
        min: int,
        negative: bool = False,
        negatives_allowed: bool = False,
    ) -> int:
        """Parses a velocity or angle value based on the minimum / maximum
        values and the stagnant values.

        Args:
            value (int): the input value.
            max (int): the maximal allowed value.
            min (int): the minimal allowed value.
            negative (bool, optional): Whether the input value should be transformed into its negative. Defaults to `False`.
            negatives_allowed (bool, optional): Whether an input value can be negative. Defaults to `False`.

        Returns:
            int: the input value so that it is smaller or equal to the maximum value and higher or equal to the minimum value if negatives are not allowed; transformed to its negative if `negative` is `True` and higher or equal to the negative of the maximal value and lower or equal to the maximal value if negatives are allowed.
        """
        if negatives_allowed:
            if value > max:
                return max
            if value < max * (-1):
                return max * (-1)
            return value

        if negative:
            if value > max:
                # velocity over max
                return max * (-1)
            if value > min:
                # velocity over min below max
                return value * (-1)
            if value > max * (-1):
                # velocity below min, over negative max
                return value
            # velocity over negative max
            return max * (-1)

        if value < min:
            return min
        if value > max:
            return max
        return value

    def _parse_velocity(
        self, velocity: int, negative: bool = False, both: bool = False
    ) -> int:
        """Parses the velocity.

        Args:
            velocity (int): the input value.
            negative (bool, optional): Whether the input value should
            be transformed into its negative. Defaults to `False`.
            both (bool, optional): Whether an input value can be negative.
            Defaults to `False`.

        Returns:
            int: the input value so that it is smaller or equal to the maximum value and higher or equal to the minimum value if negatives are not allowed; transformed to its negative if `negative` is `True` and higher or equal to the negative of the maximal value and lower or equal to the maximal value if negatives are allowed.
        """
        return self._parse_value(
            velocity,
            self.max_velocity,
            self.min_velocity,
            negative=negative,
            negatives_allowed=both,
        )

    def _parse_angle(
        self, angle: int, negative: bool = False, both: bool = False
    ) -> int:
        """Parses the angle.

        Args:
            angle (int): the input value.
            negative (bool, optional): Whether the input value should be transformed into its negative. Defaults to `False`.
            both (bool, optional): Whether an input value can be negative. Defaults to `False`.

        Returns:
            int: the input value so that it is smaller or equal to the maximum value and higher or equal to the minimum value if negatives are not allowed; transformed to its negative if `negative` is `True` and higher or equal to the negative of the maximal value and lower or equal to the maximal value if negatives are allowed.
        """
        return self._parse_value(
            angle,
            self.max_angle,
            self.min_angle,
            negative=negative,
            negatives_allowed=both,
        )

    def drive_forward(self, misty: Misty, velocity: int) -> Dict:
        """Drives forwards with the specified velocity."""
        forw = {
            "LinearVelocity": self._parse_velocity(velocity),
            "AngularVelocity": self.stagnant_angle,
        }
        return misty.perform_action("drive", data=forw)

    def drive_backward(self, misty: Misty, velocity: int):
        """Drives backwards with the specified velocity."""
        back = {
            "LinearVelocity": self._parse_velocity(velocity, negative=True),
            "AngularVelocity": self.stagnant_angle,
        }
        return misty.perform_action("drive", data=back)

    def drive_left(self, misty: Misty, velocity: int, angle: int):
        """Drives left with the specified velocity and angle."""
        left = {
            "LinearVelocity": self._parse_velocity(velocity),
            "AngularVelocity": self._parse_angle(angle),
        }
        return misty.perform_action("drive", data=left)

    def drive_right(self, misty: Misty, velocity: int, angle: int):
        """Drives right with the specified velocity and angle."""
        velocity = self._parse_velocity(velocity)
        right = {
            "LinearVelocity": self._parse_velocity(velocity),
            "AngularVelocity": self._parse_angle(angle, negative=True),
        }
        return misty.perform_action("drive", data=right)

    def stop_driving(self, misty: Misty):
        """Stops driving."""
        return misty.perform_action("drive_stop")



"""This module contains generator functions."""
import random
import string


def get_random_string(n: int) -> str:
    """Constructs an `n` characters long random string containing ASCII letters and digits.

    Args:
        n (int): The required length of the string.

    Returns:
        str: The random string.
    """

    assert n > 0, "Required string length must be a positive integer."

    return "".join(
        random.SystemRandom().choice(string.ascii_letters + string.digits)
        for _ in range(n)
    )



"""This module removes non-system files from Misty's memory."""
import os
from typing import Dict, List, Tuple

from misty2py.robot import Misty
from misty2py.utils.messages import success_parser_from_dicts
from misty2py.utils.utils import get_abs_path


def get_non_system_assets(
    misty: Misty,
    assets: List = ["audio", "image", "video", "recording"],
) -> Dict:
    """Obtains the list of all non-system assets that fall under one of the `assets` types.

    Args:
        misty (Misty): The Misty whose non system assets to obtain.
        assets (List, optional): The types of assets to obtain. Defaults to `["audio", "image", "video", "recording"]`.

    Returns:
        Dict: all non-system assets of a given type (keywords) listed as their names in the list under the type keyword.
    """
    action = "_list"
    non_sys_assets = {}
    for asset in assets:
        data = misty.get_info(asset + action)
        result = data.get("result", [])
        for hit in result:
            name = hit.get("name")
            if not name:
                continue

            is_system = hit.get("systemAsset", True)
            if not is_system:
                if not non_sys_assets.get(asset):
                    non_sys_assets[asset] = []
                non_sys_assets[asset].append(name)
                print(
                    "Found a non-system asset \
                        of type `%s` named `%s`." % (asset, name)
                )

    return non_sys_assets


def get_asset_properties(asset_type: str, file: str) -> Tuple[Dict, str, str]:
    """Obtains the parameters, the name and the extension of a file.

    Args:
        asset_type (str): The type of the asset.
        file (str): The file name.

    Returns:
        Tuple[Dict, str, str]: the parameters (a dict with keys `Name` for recordings or `FileName` for other asset types containing the name and the key `Base64` set to `"true"`), the file name and the file extension.
    """
    if asset_type == "recording":
        params = {"Name": file, "Base64": "true"}
    else:
        params = {"FileName": file, "Base64": "true"}

    split_file_name = file.split(".")
    name = split_file_name[0]
    if len(split_file_name) > 1:
        ext = split_file_name[1]
    else:
        ext = "unknown"

    return params, name, ext


def save_base64_str(
    full_path: str,
    content: str,
    overwrite: bool = False
) -> bool:
    """Saves an asset content to the file specified by `full_path`.

    Args:
        full_path (str): The absolute path to the save location.
        content (str): The base64 content to save.
        overwrite (bool, optional): Whether the overwrite the file if it exists. Defaults to `False`.

    Returns:
        bool: The successfulness of the operation.
    """
    if not overwrite and os.path.exists(full_path):
        print("File `%s` already exists, not overwriting." % full_path)
        return True

    try:
        with open(full_path, "w") as f:
            f.write(content)
        print("Asset saved into `%s`." % full_path)
        return True

    except Exception as e:
        print("Failed to save the asset into `%s`. Error message: `%s`" % (
            full_path,
            e
            )
        )
        return False


def save_assets(misty: Misty, assets: Dict, location: str) -> List[str]:
    """Saves multiple assets.

    Args:
        misty (Misty): the Misty whose assets to save.
        assets (Dict): The dictionary of assets; a list of file names under their asset type as the keyword.
        location (str): The absolute path to the directory where the assets are saved.

    Returns:
        List[str]: A list of file names for the files that failed to be saved.
    """
    failed_list = []
    action = "_file"
    for asset_type, files in assets.items():
        for file in files:
            params, name, ext = get_asset_properties(asset_type, file)
            response = misty.get_info(asset_type + action, params=params)
            result = response.get("result")

            if not result:
                failed_list.append(file)

            else:
                file_name = "%s_%s_%s_in_base64.txt" % (asset_type, name, ext)
                full_path = os.path.join(location, file_name)
                file_content = result.get("base64")
                if not file_content:
                    failed_list.append(file)
                else:
                    success = save_base64_str(full_path, file_content)
                    if not success:
                        failed_list.append(file)

    return failed_list


def delete_assets(
    misty: Misty,
    assets: Dict,
    ignore_list: List = []
) -> List[str]:
    """Deletes all non-system files except those that are in `ignore_list`.

    Args:
        misty (Misty): the Misty whose assets to delete.
        assets (Dict): The dictionary of assets; a list of file names under their asset type as the keyword.
        ignore_list (List, optional): The list of file names for files that should not be deleted. Defaults to `[]`.

    Returns:
        List[str]: The list of files that were succesfully deleted.
    """
    action = "_delete"
    delete_list = []

    for asset_type, files in assets.items():
        for file in files:
            if file not in ignore_list:
                if asset_type == "recording":
                    data = {"Name": file}
                else:
                    data = {"FileName": file}
                response = misty.perform_action(asset_type + action, data=data)
                status = response.get("status")
                if status:
                    if status == "Success":
                        print("Successfully deleted the asset `%s`." % file)
                        delete_list.append(file)
                    else:
                        print(
                            "Failed to delete the asset `%s`. Message: `%s`"
                            % (file, response)
                        )

    return delete_list


def free_memory(
    misty: Misty,
    assets: List = ["audio", "image", "video", "recording"],
    save: bool = True,
    save_dir: str = "data",
) -> Dict:
    """Removes all non-system files in Misty's memory.

    Args:
        misty (Misty): the Misty whose memory to free.
        assets (List, optional): The list of asset types to delete. Defaults to `["audio", "image", "video", "recording"]`.
        save (bool, optional): Whether to save the files that are being deleted from Misty's memory. Defaults to `True`.
        save_dir (str, optional): The path to the directory that will store saved assets. Defaults to `"data"`.

    Returns:
        Dict: The dictionary with keys `overall_success` for the success of the entire operation and keys corresponding to the actions within the operation with successfulness of the action and messages about the action.
    """
    save_dir = get_abs_path(save_dir)
    enable_audio = misty.perform_action("audio_enable")
    enable_av = misty.perform_action("streaming_av_enable")
    enable_camera = misty.perform_action("camera_enable")

    assets_to_delete = get_non_system_assets(misty, assets=assets)
    deletion = {}

    if len(assets_to_delete) == 0:
        deletion["status"] = "Success"
        deletion["message"] = "No non-system files present."

    else:
        failed_to_save_list = []

        if save:
            failed_to_save_list = save_assets(
                misty,
                assets_to_delete,
                save_dir
            )

        deleted = delete_assets(misty, assets_to_delete, failed_to_save_list)

        if len(deleted) > 0:
            deletion["status"] = "Success"
            deletion["message"] = "Successfully deleted \
                following assets: %s" % str(deleted)

        else:
            deletion["status"] = "Failed"
            deletion["message"] = "Failed to delete any assets."

    disable_audio = misty.perform_action("audio_disable")
    disable_av = misty.perform_action("streaming_av_disable")
    disable_camera = misty.perform_action("camera_disable")

    return success_parser_from_dicts(
        enable_audio=enable_audio,
        enable_av=enable_av,
        enable_camera=enable_camera,
        deletion=deletion,
        disable_audio=disable_audio,
        disable_av=disable_av,
        disable_camera=disable_camera,
    )


if __name__ == "__main__":
    from misty2py.utils.utils import get_misty

    print(free_memory(get_misty(), "data"))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['misty2py', 'misty2py.basic_skills', 'misty2py.utils']

package_data = \
{'': ['*']}

install_requires = \
['python-dotenv>=0.17.1,<0.18.0',
 'requests>=2.25.1,<3.0.0',
 'websocket-client>=0.58.0,<0.59.0']

setup_kwargs = {
    'name': 'misty2py',
    'version': '4.2.0',
    'description': 'A Python library for Misty II development',
    'long_description': '# Misty2py\n\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black) [![GitHub license](https://img.shields.io/github/license/Naereen/StrapDown.js.svg)](https://github.com/ChrisScarred/misty2py/blob/main/LICENSE)\n\nMisty2py is a Python 3 package for Misty II development using [Misty\'s REST API](https://docs.mistyrobotics.com/misty-ii/rest-api/api-reference/ "Misty Robotics REST API").\n\n## Installation\n\n### Poetry\n\nTo install misty2py, run `pip install misty2py`.\n\n### From source\n\n- If this is your first time using `misty2py` from source, do following:\n\n  - Get Poetry (`python -m pip install poetry`) if you do not have it yet.\n  - Copy `.env.example` to `.env`.\n  - Replace the placeholder values in the new `.env` file.\n  - Run `poetry install` to obtain all dependencies.\n\n- Run the desired script via `poetry run python -m [name]` where `[name]` is the placeholder for the module location (in Python notation).\n- If the scripts run but your Misty does not seem to respond, you have most likely provided an incorrect IP address for `MISTY_IP_ADDRESS` in `.env`.\n- Pytests can be run via `poetry run pytest .`.\n- The coverage report can be obtained via `poetry run pytest --cov-report html --cov=misty2py tests` for HTML output or via `poetry run pytest --cov=misty2py tests` for terminal output.\n\n## Features\n\nMisty2py can be used to develop complex skills (behaviours) for the Misty II robot utilising:\n\n- **actions** via sending a `POST` or `DELETE` requests to Misty\'s API;\n- **informations** via sending a `GET` request to Misty\'s API;\n- **continuous streams of data** via subscribing to event types on Misty\'s websockets.\n\nMisty2py uses following concepts for easy of usage:\n\n- **action keywords** - customisable python-styled keywords for endpoints of Misty\'s API that correspond to performing actions;\n- **information keywords** - customisable python-styled keywords for endpoints of Misty\'s API that correspond to retrieving information;\n- **data shortcuts** - customisable python-styled keywords for commonly used data that are supplied to Misty\'s API as the body of a `POST` request.\n\n## Usage\n\n### Getting started\n\n- Start by making **a new instance** of `misty2py.robot`\'s `Misty` by `misty_robot = Misty("ip_address_here")`\n  - Substitute `ip_address_here` with the IP address of your Misty.\n  - `misty2py.utils.env_loader` module contains `EnvLoader` class that can be used to load Misty\'s IP from the .env file in your project\'s home directory:\n    - Create the `.env` file and write `MISTY_IP_ADDRESS="[ip_address_here]"` in it.\n    - Initialise an `EnvLoader` object via `env_loader = EnvLoader()`; this loads the environment variables.\n    - Use `env_loader.get_ip()` to obtain the IP address.\n- Use the method `misty_robot.perform_action()` to tell Misty to **perform an action**.\n- Use the method `misty_robot.get_info()` to tell Misty to **return information**.\n- Use the method `misty_robot.event()` to initialise, obtain and stop **continuous streams of data** from Misty\'s event types.\n\n### Obtaining information\n\nObtaining digital information is handled by `misty2py.robot::get_info` method.\n\n`misty2py.robot::get_info` has following arguments:\n\n- `info_name` - *required;* the string information keyword corresponding to an endpoint in Misty\'s API;\n- `params` - *optional;* a dictionary of parameter name and parameter value pairs, defaults to `{}`.\n\n### Performing actions\n\nPerforming physical and digital actions including removal of non-system files is handled by `misty2py.robot::perform_action()` method.\n\n`misty2py.robot::perform_action()` has following arguments:\n\n- `action_name` - *required;* the string action keyword corresponding to an endpoint in Misty\'s API;\n- `data` - *optional;* the data to pass to the request as a dictionary or a data shortcut (string), defaults to `{}`.\n\n### Event types\n\nTo obtain event data in Misty\'s framework, it is required to **subscribe** to an event type on Misty\'s websocket server. Misty\'s websocket server then streams data to the websocket client, in this implementation via a separate thread. To **access this data,** `misty2py.robot::event` method must be called with `"get_data"` parameter from the main thread. When data are no longer required to be streamed to the client, an event type can be **unsubscribed** to kill the event thread.\n\n#### Subscription\n\nSubscribe to an event via `misty2py.robot::event` with the parameter `"subscribe"` and following keyword arguments:\n\n    - `type` - *required;* event type string as documented in [Event Types Docs](https://docs.mistyrobotics.com/misty-ii/robot/sensor-data/ "Misty Robotics Event Types").\n    - `name` - *optional;* a custom event name string; must be unique.\n    - `return_property` - *optional;* the property to return from Misty\'s websockets; all properties are returned if return_property is not supplied.\n    - `debounce` - *optional;* the interval in ms at which new information is sent; defaults to 250.\n    - `len_data_entries` - *optional;* the maximum number of data entries to keep (discards in fifo style); defaults to 10.\n    - `event_emitter` - *optional;* an event emitter function which emits an event upon message recieval. Supplies the message content as an argument.\n\n#### Accessing the data and the log\n\nAccess the data of an event or its log via `misty2py.robot::event` with the parameter `"get_data"` or `"get_log"` and a keyword argument `name` (the name of the event).\n\n#### Unsubscribing\n\nUnsubscribe from an event via `misty2py.robot::event` with the parameter `"unsubscribe"` and a keyword argument `name` (the name of the event).\n\n#### Basic example\n\n```python\nimport time\n\nfrom misty2py.robot import Misty\nfrom misty2py.utils.env_loader import EnvLoader\n\nenv_loader = EnvLoader\n\nm = Misty(env_loader.get_ip())\n\nd = m.event("subscribe", type = "BatteryCharge")\ne_name = d.get("event_name")\n\ntime.sleep(1)\n\nd = m.event("get_data", name = e_name)\n\nd = m.event("unsubscribe", name = e_name)\n```\n\n#### Event emitter usage - example\n\n```python\nimport time\nfrom pymitter import EventEmitter\n\nfrom misty2py.robot import Misty\nfrom misty2py.utils.env_loader import EnvLoader\n\nenv_loader = EnvLoader\n\nm = Misty(env_loader.get_ip())\nee = EventEmitter()\nevent_name = "myevent_001"\n\n@ee.on(event_name)\ndef listener(data):\n    print(data)\n\nd = m.event("subscribe", type = "BatteryCharge", name = event_name, event_emitter = ee)\n\ntime.sleep(2)\n\nd = m.event("unsubscribe", name = event_name)\n```\n\n### Adding custom keywords and shortcuts\n\nCustom keywords and shortcuts can be passed to a Misty object while declaring a new instance by using the optional arguments:\n\n- `custom_info` for custom information keywords (a dictionary with keys being the information keywords and values being the endpoints),\n- `custom_actions` for custom action keywords (a dictionary with keys being the action keywords and values being a dictionary `{"endpoint" : "edpoint_value", "method" : "method_value"}` where `method_value` is either `post` or `delete`),\n- `custom_data` for custom data shortcuts (a dictionary with keys being the data shortcuts and values being the dictionary of data values).\n\nAn example:\n\n```python\ncustom_allowed_infos = {\n    "hazards_settings": "api/hazards/settings"\n}\n\ncustom_allowed_data = {\n    "amazement": {\n        "FileName": "s_Amazement.wav"\n    },\n    "red": {\n        "red": "255",\n        "green": "0",\n        "blue": "0"\n    }\n}\n\ncustom_allowed_actions = {\n    "audio_play" : {\n        "endpoint" : "api/audio/play",\n        "method" : "post"\n    },\n    "delete_audio" : {\n        "endpoint" : "api/audio",\n        "method" : "delete"\n    }\n}\n\nmisty_robot = Misty("0.0.0.0", \n    custom_info=custom_allowed_infos, \n    custom_actions=custom_allowed_actions, \n    custom_data=custom_allowed_data)\n```\n',
    'author': "Chris 'Scar(R)ed' Vajdík",
    'author_email': 'scarred.chris@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ChrisScarred/misty2py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""This module enables setting an expression via a single function.
"""

import time
from typing import Dict, Union

from misty2py.robot import Misty
from misty2py.utils.colours import (construct_transition_dict,
                                    get_rgb_from_unknown)
from misty2py.utils.messages import success_parser_from_list
from misty2py.utils.status import ActionLog


def expression(
    misty: Misty,
    image: Union[str, Dict, None] = None,
    sound: Union[str, Dict, None] = None,
    colour_type: Union[str, None] = None,
    colour: Union[str, Dict, None] = None,
    colours: Union[str, Dict, None] = None,
    image_offset: Union[float, int] = 0,
    sound_offset: Union[float, int] = 0,
    colour_offset: Union[float, int] = 0,
    duration: Union[float, int] = 1.5,
) -> Dict:
    """Performs an audio-visual expression. 

    Args:
        misty (Misty): the Misty that performs the expression.
        image (Union[str, Dict, None], optional): The path to the image to display or `None` if no special image should be displayed. Defaults to `None`.
        sound (Union[str, Dict, None], optional): The path to the sound to play or None if no sound should be played. Defaults to `None`.
        colour_type (Union[str, None], optional): `"trans"` if a colour transition should be lit. Defaults to `None`.
        colour (Union[str, Dict, None], optional): The colour of LED as a dictionary or a data shortcut or `None` if no special colour should be lit. Defaults to `None`.
        colours (Union[str, Dict, None], optional): The LED colours as a dictionary or a data shortcut or `None` if no transitioning colours should be lit. Defaults to `None`.
        image_offset (Union[float, int], optional): The offset between starting the expression and displaying the image in ms. Defaults to `0`.
        sound_offset (Union[float, int], optional): The offset between starting the expression and playing the sound in ms. Defaults to `0`.
        colour_offset (Union[float, int], optional): The offset between starting the expression and lighting the LEDs in ms. Defaults to `0`.
        duration (Union[float, int], optional): The duration of the expression. Defaults to `1.5`.

    Returns:
        Dict: The responses from Misty's REST API for every action and the key `overal_success` (bool) specifying whether all actions were successful.
    """
    assert duration > 0, "Duration must be higher than zero."
    assert (
        colour_offset >= 0 and sound_offset >= 0 and image_offset >= 0
    ), "Offsets must be higher or equal to zero"
    assert (
        image or sound or colour or colours
    ), "At least one audio-visual component \
        (display image, sound or led colour / colours) must be set."

    assert (
        image_offset < duration and
        sound_offset < duration and
        colour_offset < duration
    ), "The offsets cannot be higher than the duration."

    actions = ActionLog()

    reset_img = False
    reset_led = False

    offsets = sorted(set([image_offset, sound_offset, colour_offset]))
    for i, offset in enumerate(offsets):
        time.sleep(offset)

        if offset == image_offset:
            if image:
                img_show_message = misty.perform_action(
                    "image_show",
                    data=image
                )
                actions.append_({"img_show": img_show_message})
                reset_img = True

        if offset == sound_offset:
            if sound:
                audio_play_message = misty.perform_action(
                    "audio_play",
                    data=sound
                )
                actions.append_({"audio_play": audio_play_message})

        if offset == colour_offset:
            if colour_type == "trans":
                json_cols = None
                if colours:
                    json_cols = construct_transition_dict(
                        colours,
                        misty.actions.allowed_data
                    )
                elif colour:
                    json_cols = construct_transition_dict(
                        {"col1": colour, "col2": "led_off"},
                        misty.actions.allowed_data
                    )
                if json_cols:
                    led_trans_message = misty.perform_action(
                        "led_trans",
                        data=json_cols
                    )
                    actions.append_({"led_trans": led_trans_message})
                    reset_led = True
            else:
                if colour:
                    json_col = get_rgb_from_unknown(
                        colour, allowed_data=misty.actions.allowed_data
                    )
                    led_message = misty.perform_action("led", data=json_col)
                    actions.append_({"led": led_message})
                    reset_led = True

        if i == len(offsets) - 1:
            time.sleep(duration - offset)

    if reset_img:
        reset_img_message = misty.perform_action(
            "image_show", data="image_content_default"
        )
        actions.append_({"reset_img": reset_img_message})

    if reset_led:
        reset_led_message = misty.perform_action("led", data="led_off")
        actions.append_({"reset_led": reset_led_message})

    return success_parser_from_list(actions.get_())
